/* Licensed under the GPLv2 */

/* Originally written by marek@bmlv.gv.at 2000 */
/* make it work with pcsclite-1.0.1: Vincent Guyot <vguyot@inf.enst.fr>  2002-07-12 */


#include <stdio.h>
#include <winscard.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>


#define MAXBUFF (512)



/* Globals */

SCARDCONTEXT g_sc_context=0;
SCARDHANDLE g_card_hdl=0;


char g_buffer_i[MAXBUFF];
char g_buffer_o[MAXBUFF];


/* Options */

char
  *o_readername=NULL;

int 
  o_explanation=0,
  o_stdprotocol=SCARD_PROTOCOL_T0,
  o_shared=0,
  o_transaction=1;


/* Error "procedure" */

#define E(x)  \
  if (x!=SCARD_S_SUCCESS)  \
  {  \
    int e=x; \
    OnExit(); \
    fprintf(stderr,"Err: %s#%d => %d=0x%X\n",__FILE__,__LINE__,e,e);  \
    sprintf(g_buffer_i,"0x%04x",e); \
    execlp("grep",  \
        "grep","-i",g_buffer_i,"/usr/local/include/pcsclite.h",NULL); \
    exit(1);  \
  }


/* explanatory texts */

typedef struct { char msk[2], rsp[2], *text; } t_response;

const t_response C_response[]=
    {
      { { 0xff, 0xff }, { 0x90, 0x00 } , "Ok" },

      { { 0xff, 0x00 }, { 0x61, 0x00 } , "%d response bytes available" },
      { { 0xff, 0x00 }, { 0x9f, 0x00 } , "%d response bytes available" },

      { { 0xff, 0xff }, { 0x62, 0x81 } , "returned data may be corrupt" },
      { { 0xff, 0xff }, { 0x62, 0x82 } , "EOF reached prematurely" },
      { { 0xff, 0xff }, { 0x62, 0x83 } , "selected file invalid" },
      { { 0xff, 0xff }, { 0x62, 0x84 } , "FCI not formated" },
      { { 0xff, 0x00 }, { 0x62, 0x00 } , "nvmem unchanged" },

      { { 0xff, 0x00 }, { 0x63, 0x00 } , "nvmem changed" },
      { { 0xff, 0x00 }, { 0x63, 0x81 } , "file filled up by last write" },
      { { 0xff, 0xf0 }, { 0x63, 0xc0 } , "Counter=%1.1X" },

      { { 0xff, 0xff }, { 0x64, 0x00 } , "nvmem unchanged" },
      { { 0xff, 0x00 }, { 0x64, 0x00 } , "nvmem unchanged - RFU" },

      { { 0xff, 0xff }, { 0x65, 0x00 } , "nvmem changed" },
      { { 0xff, 0xff }, { 0x65, 0x81 } , "nvmem changed - memory failure" },
      { { 0xff, 0x00 }, { 0x65, 0x00 } , "nvmem changed - unknown?" },

      { { 0xff, 0x00 }, { 0x66, 0x00 } , "security related %d" },

      { { 0xff, 0xff }, { 0x67, 0x00 } , "wrong length" },
      { { 0xff, 0x00 }, { 0x67, 0x00 } , "wrong length - %d expected" },

      { { 0xff, 0xff }, { 0x68, 0x81 } , "wrong cla - logical channel not supported" },
      { { 0xff, 0xff }, { 0x68, 0x82 } , "wrong cla - secure messaging not supported" },
      { { 0xff, 0x00 }, { 0x68, 0x00 } , "cla not supported" },

      { { 0xff, 0xff }, { 0x69, 0x81 } , "command incompatible with file structure" },
      { { 0xff, 0xff }, { 0x69, 0x82 } , "security status not satisfied" },
      { { 0xff, 0xff }, { 0x69, 0x83 } , "authentication method blocked" },
      { { 0xff, 0xff }, { 0x69, 0x84 } , "referenced data invalid" },
      { { 0xff, 0xff }, { 0x69, 0x85 } , "conditions of use not satisfied" },
      { { 0xff, 0xff }, { 0x69, 0x86 } , "command not allowed - no current EF" },
      { { 0xff, 0xff }, { 0x69, 0x87 } , "expected SM data objects missing" },
      { { 0xff, 0xff }, { 0x69, 0x88 } , "SM data objects incorrect" },
      { { 0xff, 0x00 }, { 0x69, 0x00 } , "command not allowed" },

      { { 0xff, 0xff }, { 0x6a, 0x80 } , "P1-P2: incorrect parameters in data field" },
      { { 0xff, 0xff }, { 0x6a, 0x81 } , "P1-P2: function not supported" },
      { { 0xff, 0xff }, { 0x6a, 0x82 } , "P1-P2: file not found" },
      { { 0xff, 0xff }, { 0x6a, 0x83 } , "P1-P2: record not found" },
      { { 0xff, 0xff }, { 0x6a, 0x84 } , "P1-P2: not enough memory space in file" },
      { { 0xff, 0xff }, { 0x6a, 0x85 } , "P1-P2: Lc inconsistent with TLV" },
      { { 0xff, 0xff }, { 0x6a, 0x86 } , "P1-P2 incorrect" },
      { { 0xff, 0xff }, { 0x6a, 0x87 } , "P1-P2 inconsistent with Lc" },
      { { 0xff, 0xff }, { 0x6a, 0x88 } , "Referenced data not found" },
      { { 0xff, 0x00 }, { 0x6a, 0x00 } , "P1-P2 invalid" },

      { { 0xff, 0x00 }, { 0x6b, 0x00 } , "P1-P2 invalid" },

      { { 0xff, 0x00 }, { 0x6c, 0x00 } , "wrong length -  %d expected" },

      { { 0xff, 0x00 }, { 0x6d, 0x00 } , "INS code not supported or invalid" },
      
      { { 0xff, 0x00 }, { 0x6e, 0x00 } , "CLA %02X not supported" },

      { { 0xff, 0x00 }, { 0x6f, 0x00 } , "no precise diagnosis" },

      { { 0x00, 0x00 }, { 0x00, 0x00 } , "Unknown response" }
    };




/****************** CLOSE **************/  

void OnExit(void) __attribute__ ((destructor));
void OnExit(void)
{
  LONG ret;

  if (g_card_hdl)
  {
//    ret=SCardDisconnect(g_card_hdl,SCARD_RESET_CARD);
    ret=SCardDisconnect(g_card_hdl,SCARD_UNPOWER_CARD);
    g_card_hdl=0;
    E(ret);
  }

  if (g_sc_context)
  {
    ret=SCardReleaseContext(g_sc_context); 
    g_sc_context=0;
    E(ret);
  }
}


/****************** SUB **************/  

int Interactive(void)
{
  static SCARD_IO_REQUEST scir;
  static char g_getresponse[5]= {0xa0,0xc0,0x00,0x00 };
  int c, l,j,i,p;
  LONG ret,len,prot,len2;
  char *cp;


  if (feof(stdin)) return 0;

  if (!fgets(g_buffer_i,MAXBUFF,stdin))
    return 1;

  if ((len2=strlen(g_buffer_i)) && isspace(g_buffer_i[len2]))
    g_buffer_i[len2]=0;

  cp=g_buffer_i;
  len2=0;
  while (*cp && sscanf(cp,"%02X %n",&c,&l)>0)
  {
    g_buffer_o[len2]=c;
    len2++;
    cp+=l;
  }

  if (len2==0) return 1;


  len=MAXBUFF;
  ret=SCardTransmit(g_card_hdl,
     // prot ? SCARD_PCI_T1 : SCARD_PCI_T0,
      prot==SCARD_PROTOCOL_T1 ? SCARD_PCI_T1 : SCARD_PCI_T0,
      g_buffer_o,len2,&scir,
      (BYTE *)g_buffer_i,&len);
  E(ret);

  if (len==2)
  {
    switch ((unsigned char)g_buffer_i[0])
    {
      case 0x61:
      case 0x9f:
        if (g_buffer_i[1]==0) 
        {
          break;
        }
        g_getresponse[4]=g_buffer_i[1];

        len=MAXBUFF;
        ret=SCardTransmit(g_card_hdl,
           // prot ? SCARD_PCI_T1 : SCARD_PCI_T0,
            prot==SCARD_PROTOCOL_T1 ? SCARD_PCI_T1 : SCARD_PCI_T0,
            g_getresponse,sizeof(g_getresponse),&scir,
            (BYTE *)g_buffer_i,&len);
        E(ret);
        break;
    }
  } 

  for(j=0; j<len; j++)
    printf("%02X",g_buffer_i[j] & 0xff);

  if (o_explanation)
  {
    fputs("  ",stdout);

    if (len>=2)
    {
      p=len-2;
      for(j=0; j< sizeof(C_response)/sizeof(C_response[0]); j++)
      {
	for(i=sizeof(C_response[j].msk)-1; i>=0; i--)
	  if (C_response[j].rsp[i] != (C_response[j].msk[i] & g_buffer_i[p+i]))
	    break;

	if (i==-1)
	{
	  printf(C_response[j].text,(int)(unsigned char)g_buffer_i[p+1]);
	  break;
	}
      }
    }		// if len>=2
  }		// if o_explanation

  return 1;
}

/****************** MAIN **************/  

int main(int argc, char *args[])
{
  int opt;
  LONG ret,len,prot;

  while ((opt=getopt(argc,args,"h?er:p:tVs")) != -1)
  {
    switch (opt)
    {
      case 'r':
	o_readername=optarg;
	break;
      case 'p':
	o_stdprotocol=atoi(optarg);
	break;
      case 's':
	o_shared^=1;
	break;
      case 't':
	o_transaction^=1;
	break;
      case 'V':
	puts("cardcomm, version 0.99 as of 2000.05.08");
	exit(0);
	break;

      case 'e':
	o_explanation^=1;
	break;

      case 'h':
      case '?':
      default:
	puts(
	    "Usage: [-e] [-t] [-rName] [-pProt#] [-s]\n"
	    "-e\texplanations on/OFF\n"
	    "-s\tshared access on/OFF\n"
	    "-t\tmake it a transaction ON/off\n"
	    "-rName\treader to connect to\n"
	    "-pProt\tdefault protocol number as given in <pcsclite.h>\n"
	    "\tas time of compile: T0=1, T1=2, RAW=4\n\n"
	    "Usage: -V\n"
	    "-V\tprint version and exit\n"
	    "");
	exit(1);
    }
  }
  
  
  signal(SIGPIPE,(void *)(int)OnExit);

/****************** INIT **************/  

  ret=SCardEstablishContext(SCARD_SCOPE_GLOBAL,NULL,NULL,&g_sc_context); 
  E(ret);

  if (!o_readername)
  {
    //  len=17;
    ret=SCardListReaders(g_sc_context,NULL,NULL,&len);
    E(ret);

    if (len>MAXBUFF) exit(5);

    ret=SCardListReaders(g_sc_context,NULL,g_buffer_i,&len);
    E(ret);

    o_readername=g_buffer_i;
  }
 
  while (1)
  {
    ret=SCardConnect(g_sc_context,o_readername,
        SCARD_SHARE_SHARED,
//        SCARD_SHARE_EXCLUSIVE,
        o_stdprotocol,
      &g_card_hdl,&prot);

    if (ret == SCARD_S_SUCCESS) break;

    if (ret == SCARD_E_NO_SMARTCARD)
    {
      sleep(2);
      continue;
    }

    E(ret);
  }
 

/****************** WORK **************/  

  if (o_transaction)
  {
    ret=SCardBeginTransaction(g_card_hdl);
    E(ret);
  }

  while (Interactive())
  {
    /* EOL and flushing ... */
    printf("\n");
    fflush(stdout);
  }


  if (o_transaction)
  {
    ret=SCardEndTransaction(g_card_hdl,SCARD_LEAVE_CARD);
    E(ret);
  }

/****************** finish **************/  

  OnExit();
  return 0;
}

