#!/usr/bin/perl

require "abbrev.pl";
use IPC::Open2;


$PIN="";

$ENV{"COM1"}="/dev/ttyS0";


%cmds= (
  "help" 	=> [ \&Help, 
                     "gives help", "" ],
  "forget" 	=> [ \&Forget, 
                     "forgets PIN in memory", "" ],
  "chgpin" 	=> [ \&ChangePIN, 
                     "changes pin.", "{new pin {1|2}}" ],
  "givepin" 	=> [ \&PresentPIN, 
                     "gives pin to card.", "{pin {1|2}}" ],
  "opencard" 	=> [ \&OpenCard, 
                     "opens card", "{pin}" ],
  "closecard" 	=> [ \&CloseCard, 
                     "closes card.", "" ],
  "reqpin1"	=> [ \&EnDisablePIN, 
                     "en/disable pin1", "{0|1} {pin1}" ],
  "unblockpin"	=> [ \&UnblockPIN, 
                     "unblocks pin.", "{1|2 {new pin {puk}}}" ],
  "status"	=> [ \&Status, 
                     "print status of card & connection", "" ],
  "info"	=> [ \&FileInfo, 
                     "various info.", "{serial|lang|imsi|sp}" ],
  "write"	=> [ \&WritePhoneBook, 
                     "writes phonebook.", "{filename {empty}}" ],
  "read"	=> [ \&ReadPhoneBook, 
                     "reads phonebook.", "{filename {empty}}" ],
  "quit"	=> [ \&Quit, "exit program" ],
);

%FileInfoIDs=(
    "serial" => [ "card serial number", [ "2fe2" ],
      sub { unpack("h*",$bin) =~ /^([0-9]+)/; $1; } ],
    "lang" => [ "language preference code(s) ", [ "7f20", "6f05" ],
      sub { join(", ",grep($_ && $_ != 255,unpack("C*",$bin))) } ],
    "imsi" => [ "intern. mob. subscr. id", [ "7f20", "6f07" ],
      sub { unpack("h*",$bin) =~ /^..([0-9]+)/; $1; } ],
    "sp" => [ "service provider name", [ "7f20", "6f46" ],
      sub { $bin =~ /^.([^\xff]*)/; $1 || "(none)"; } ],
  );

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 

%cmdabbrev=();
&abbrev(*cmdabbrev,keys %cmds);

#print map { $_, " => ",$cmdabbrev{$_},"\n" } keys %cmdabbrev;



$Quit=0;
while (!$Quit)
{
  print "Enter Command: ";
  last unless $_=scalar(<STDIN>);

  s/^\s+//;
  s/\s+$//;

  next if $_ eq "";

  ($cmd,@par)=split(/\s+/);

  if ($cmdref=&GetCmdRef(lc $cmd))
  {
    print &{$cmdref->[0]}(@par);
  }
  else
  {
    print qq#400 "$cmd" is not defined. try "help".\n#;
  }
}


###################

sub GetCmdRef
{
  my($ab)=@_;
  my($cmd);

  ($cmd=$cmdabbrev{$ab}) ?
    wantarray ? ($cmd,$cmds{$cmd}) : $cmds{$cmd} : undef;
}

###################

sub UnPack
{
  my($out,$template)=@_;
  my($bin);

  $bin=pack("H*",$out);

  wantarray() ?
    $template ? (unpack($template,$bin),$out) : ($bin,$out) :
    $out;
}

sub DoAPDU
{
  my($input,$template)=@_;
  my($out,$bin);

  $out=$input . "\n";
  syswrite(CARDW,$out,length($out));
  if (defined(sysread(CARDR,$out,4096)))
  {
    chomp $out;


    &UnPack($out,$template);
#$bin=pack("H*",$out);
#wantarray() ?
#$template ? (unpack($template,$bin),$out) : ($bin,$out) :
#$out;
  }
  else
  {
    die "error processing APDU $input.\n";
  }
}


sub SelectGSMFile
{
  my(@list)=@_;
  my($i,$erg);

  for $i ("3f 00",@list)
  {
    $erg=&DoAPDU("A0 A4 00 00 02 " . $i);
    return "402 select $i: $erg\n"
      unless $erg =~ /9000$/;
  }

  return wantarray ? (0,$erg) : 0;
}

###################


sub Quit
{
  &CloseCard();
  $Quit=1;
  "200 goodbye.\n";
}


sub Help
{
  my(@list)=@_;
  my($i,$cmd,$ref,$desc,$para);
#  local($~)="Help";


  print "201 help follows.\n";

  if (@list)
  {
    for $i (@list)
    {
      ($cmd,$ref)=&GetCmdRef(lc $i);

      if ($cmd)
      {
	($desc,$para)= @{$ref}[1,2];
	print "Command: $cmd\n  Parameter: $para\n  Description: $desc\n";
#      write;
      }
      else
      {
	print qq#"$i" is not defined.\n#;
      }
    }
  }
  else
  {
    print join(" ",sort keys %cmds),"\n";
  }

  ".\n";
}

format Help=
Command: @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$cmd . "."
  Parameter: @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$para
  Description: ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$desc
~~  ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$desc
.


sub CloseCard
{
  my($cnt);
  return "400 card not allocated"
    unless $cardpid;

  close CARDW;
  for($cnt=0; $cnt<1024 && scalar(<CARDR>); $cnt++) {;} 
  close CARDR;

#sleep 2;
#kill $cardpid if kill 0,$cardpid;
  wait;

  $cardpid=0;
  "200 card freed.\n";
}


sub OpenCard
{
  my($pin)=@_;

  if ($cardpid)
  {
    print "201 closing old connection ...\n",
      &CloseCard();
  }

  $cardpid=open2(\*CARDR,\*CARDW,"cardcomm/cardcomm") || die $!;

  if ($pin)
  {
    print "201 checking pin ...\n",
      &PresentPIN($pin,1);
  }

  $cardpid && !$? ? 
    "200 card allocated.\n" :
    "400 can't use card\n";
}


sub ReadPhoneBook
{
  my($file,$empty)=@_;
  my($i,$reclen,$size,$ascii);
  my($name,$len,$ssc,$ton,$npi,$number,$cci,$ext,$erg);
  local(*FILE);


  last unless $file||=&ReqInput("filename, '-' = stdout");
  $empty||=&ReqInput("empty records ? (0|1)")
      unless length($empty)!=0;

  open(FILE, ($file eq "-" ? ">& STDOUT" : "> " . $file)) 
    || return "400 open file: $!";

  ($erg,$i)=SelectGSMFile("7f10","6f3a");
  return $erg if $erg;

#  ($rfu,$size,$fileid,$type,$cyclic_increase,
#    $access,$status,$len,$struct,$reclen,$erg) =
  ($i,$size,$i,$i,$i,$i,$i,$i,$i,$reclen,$erg) =
    &UnPack($i,"n n n C C H6 C C C C");


  $ascii=$reclen - (1+1+12);

  $template="a$ascii h1 X H1 h1 X H1 h20 C C";

  $size=$size/$reclen;
  print "201 phonebook has a maximum of ",
    "$size entries  $reclen bytes.\n";
  for($i=1; $i<=$size; $i++)
  {
    ($name,$len,$ssc,$ton,$npi,$number,$cci,$ext,$erg)=
      &DoAPDU(sprintf("A0 b2 %02X 04 %02X",$i,$reclen),$template);

    return "402 read record $i: $erg" 
      unless $erg =~ /9000$/;

    $number =~ s/f+//g;
    $number =~ tr/ab/*#/;
    $number =~ s/e(.)/<expansion $1>/g;
    $number="+" . $number if $npi & 1;

    $name =~ s/\xff+//g;

    next if $name eq "" && ! $empty;
    print FILE join("\t", $i,$name,$number),"\n";
  }

  close FILE;

  ".\n";
}

sub WritePhoneBook
{
  my($file,$empty)=@_;
  my($i,$reclen,$size,$ascii,$cmd);
  my($name,$len,$ssc,$ton,$npi,$number,$cci,$ext,$erg);
  local(*FILE);


  last unless $file||=&ReqInput("filename, '-' = stdin");
  $empty=&ReqInput("write empty records ? (0|1)")
    unless length($empty)!=0;

  open(FILE, ($file eq "-" ? "<& STDIN" : "< " . $file)) 
    || return "400 open file: $!";

  ($erg,$i)=SelectGSMFile("7f10","6f3a");
  return $erg if $erg;

#  ($rfu,$size,$fileid,$type,$cyclic_increase,
#    $access,$status,$len,$struct,$reclen,$erg) =
  ($i,$size,$i,$i,$i,$i,$i,$i,$i,$reclen,$erg) =
    &UnPack($i,"n n n C C H6 C C C C");


  $ascii=$reclen - (1+1+12);


  $size=$size/$reclen;
  print "201 phonebook has a maximum of ",
    "$size entries  $reclen bytes.\n";

  $erg="200 ok.\n";   
  while (<FILE>) 
  {
    chomp;
    last if /^\s*\./;	# Punkt am Anfang = Ende

    ($i,$name,$number)=split(/\t/);
    if ($name eq "")
    {
      next if ! $empty;

      $cmd="ff" x $reclen;	# Sonderfall fr leere Records
    }
    else
    {
      $erg="400 name too long\n",last 
	if length($name)>$ascii;

      $erg="400 number too long\n",last 
	if length($number)>20;

      $cci=$ext="ff";
      $sscton="81";

      $sscton="91" if $number =~ s/^\++//;
      $number =~ tr/*#/ab/;

      $erg="400 number invalid\n",last 
	if $number =~ /[^0-9#*ab]/;

      $len=sprintf("%02X",(length($number)+1)/2);
      $number =~ s/(.)(.)/$2$1/g;

#  $template="a$ascii H2 H2 h20 C C";
      $cmd=unpack("H*",$name) . ("ff" x ($ascii - length($name))) . 
	 $len . $sscton .
	 $number . ("f" x (20 - length($number)));
    }

#    print "--> ",$cmd,"\n";

    $cmd=&DoAPDU(sprintf("A0 dc %02X 04 %02X %s",$i,$reclen,$cmd));

#    print "<-- ",$cmd,"\n";

    return "402 write record $i: $cmd" 
      unless $cmd =~ /9000$/;

  }

  close FILE;

  $erg;
}


sub Status_PINText
{
  my($i)=@_;

  ($i & 0x80 ? "" : "not "),"initizialized, ",
    ($i & 0x0f ? ($i & 0x0f) . " retries left" : "BLOCKED");
}

sub Status
{
  my($erg,$i,$raw);
  my($mem,$id,$rfu,$type,$len,@char);

  return "400 card not opened!\n"
    unless $cardpid;
  
  ($erg,$i)=SelectGSMFile();
  return $erg if $erg;

  ($rfu,$mem,$id,$type,$rfu,$len,@char)=
    &UnPack($i,"n n n C a5 C C*");

  $i =~ s/\s+//;
#1 while $i =~ s/(\S{8})(\S)/$1  $2/;
  1 while $i =~ s/(\S{4})(\S)/$1 $2/;

  print "201 card status follows\n",
    "free mem ",$mem,"\n",
    "PIN1 ",($char[0] & 0x80 ? "not " : ""),"required\n",
    "number of PINs ",$char[3],"\n",
    "PIN1 ",&Status_PINText($char[5]),"\n",
    "PUK1 ",&Status_PINText($char[6]),"\n",
    "PIN2 ",&Status_PINText($char[7]),"\n",
    "PUK2 ",&Status_PINText($char[8]),"\n",
    "\n",
    "raw data ",$i,"\n";

  ".\n";
}


sub FileInfo
{
  my(@which)=@_;
  local($i,$bin,$txt,$sub,$path,$erg,$descr,$len);

  @which=split(/\s+/,&ReqInput("info name(s)"))
    unless @which;
  
  for $i (grep(/\S/,@which))
  {
    ($descr,$path,$sub)=@{$FileInfoIDs{$i}};
    return "404 unknown: $i\n"
      unless $descr && $sub && $path;

    ($erg,$txt)=&SelectGSMFile(@$path);
    return $erg . " while getting " . $i
      if $erg; 

    $txt=&DoAPDU("a0 b0 00 00 00");
    if ($txt =~ /^67([0-9a-f]{2})$/i)
    {
      $txt=&DoAPDU("a0 b0 00 00 " . $1);
    }

    return "404 can't read " . $descr . ":" . $txt . "\n"
      unless $txt =~ s/9000$//;

    $bin=pack("H*",$txt);
    print "201 ",$descr," ",&$sub(),"\n";
  }

  return ".\n";
}


sub Forget
{
  $PIN=" " x length($PIN);
  $PIN="";
  "200 ok.";
}

###################################

sub EnDisablePIN
{
  my($enable,$pin)=@_;

  $enable||=&ReqInput("enable=1, disable=0: ")
    if length($enable)==0;
  last unless $pin||=$PIN || &ReqInput("pin1");

  $erg=&DoAPDU(sprintf("a0%s000108%16s",
	$enable ? "28" : "26",
	unpack("H16",$pin . ("\xff" x 8))
     ));

  &Choose($erg, qq#unknown response "$erg"\n#,
    "9000", "200 pin1 now " . ($enable ? "" : "not ") . "required\n");
}


sub ChangePIN
{
  my($which,$new)=@_;
  my($old);

  $which=1 unless $which==2;
  last unless $old= ($which==1 ? $PIN : 0) || &ReqInput("old pin$which");
  last unless $new||=&ReqInput("new pin$which");

  $erg=&DoAPDU(sprintf("a024000%1d10%16s%16s",
	$which,
	unpack("H16",$old . ("\xff" x 8)),
	unpack("H16",$new . ("\xff" x 8))
     ));

  $PIN=$new if $erg eq "9000" && $which==1;

  &Choose($erg, qq#unknown response "$erg"\n#,
    "9000", "200 pin$which changed\n",
    "9804", "402 pin$which wrong\n",
    "9840", "402 pin$which BLOCKED\n");
}


sub UnblockPIN
{
  my($which,$new,$puk)=@_;

  $which=1 unless $which==2;	# nur 0 oder 2 gltig
  last unless $new||=&ReqInput("new pin$which");
  last unless $puk||=&ReqInput("puk$which");

  $erg=&DoAPDU(sprintf("a02c000%1d10%16s%16s",
	$which == 2 ? 2 : 0,
	unpack("H16",$puk . ("\xff" x 8)),
	unpack("H16",$new . ("\xff" x 8))
     ));

  $PIN=$new if $erg eq "9000" && $which==1;

  &Choose($erg, qq#unknown response "$erg"\n#,
    "9000", "200 pin$which changed\n",
    "9804", "402 pin$which wrong\n",
    "9840", "402 pin$which BLOCKED\n");
}


sub PresentPIN
{
  my($pin,$which)=@_;
  my($erg);

  $which=1 unless $which==1 || $which==2;

  last unless $pin||=&ReqInput("pin$which");
  $PIN=$pin;
  $erg=&DoAPDU(sprintf("a020000%1d08%16s",
	$which,unpack("H16",$pin . ("\xff" x 8))));

  &Choose($erg, qq#unknown response "$erg"\n#,
    "9000", "200 pin$which accepted\n",
    "9804", "402 pin$which wrong\n",
    "9840", "402 pin$which BLOCKED\n")
}
 
###################### UI-stuff

sub Choose
{
  my($key,$default,%msgs)=@_;

  $msgs{$key} ? $msgs{$key} : $default;
}

sub ReqInput
{
  my($out)=@_;
  my($in);

  print "201 enter ",$out,"\n";
  $in=scalar(<STDIN>);
  chomp $in;
  $in;
}

